#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import logging
import os
import sys
from pprint import pformat

import yaml

from ludwig.data.postprocessing import postprocess
from ludwig.data.preprocessing import preprocess_for_training
from ludwig.globals import LUDWIG_VERSION, set_on_master, is_on_master
from ludwig.globals import TRAIN_SET_METADATA_FILE_NAME
from ludwig.models.modules.measure_modules import get_best_function
from ludwig.predict import predict
from ludwig.predict import print_prediction_results
from ludwig.predict import save_prediction_outputs
from ludwig.predict import save_prediction_statistics
from ludwig.train import get_experiment_dir_name
from ludwig.train import get_file_names
from ludwig.train import train
from ludwig.train import update_model_definition_with_metadata
from ludwig.utils.data_utils import save_json
from ludwig.utils.defaults import default_random_seed
from ludwig.utils.defaults import merge_with_defaults
from ludwig.utils.misc import get_experiment_description
from ludwig.utils.print_utils import logging_level_registry
from ludwig.utils.print_utils import print_ludwig


def experiment(
        model_definition,
        model_definition_file=None,
        data_csv=None,
        data_train_csv=None,
        data_validation_csv=None,
        data_test_csv=None,
        data_hdf5=None,
        data_train_hdf5=None,
        data_validation_hdf5=None,
        data_test_hdf5=None,
        train_set_metadata_json=None,
        experiment_name='experiment',
        model_name='run',
        model_load_path=None,
        model_resume_path=None,
        skip_save_progress_weights=False,
        skip_save_processed_input=False,
        skip_save_unprocessed_output=False,
        output_directory='results',
        gpus=None,
        gpu_fraction=1.0,
        use_horovod=False,
        random_seed=default_random_seed,
        debug=False,
        **kwargs
):
    """Trains a model on a dataset's training and validation splits and
    uses it to predict on the test split.
    It saves the trained model and the statistics of training and testing.
    :param model_definition: Model definition which defines the different
           parameters of the model, features, preprocessing and training.
    :type model_definition: Dictionary
    :param model_definition_file: The file that specifies the model definition.
           It is a yaml file.
    :type model_definition_file: filepath (str)
    :param data_csv: A CSV file contanining the input data which is used to
           train, validate and test a model. The CSV either contains a
           split column or will be split.
    :type data_csv: filepath (str)
    :param data_train_csv: A CSV file contanining the input data which is used
           to train a model.
    :type data_train_csv: filepath (str)
    :param data_validation_csv: A CSV file contanining the input data which is used
           to validate a model..
    :type data_validation_csv: filepath (str)
    :param data_test_csv: A CSV file contanining the input data which is used
           to test a model.
    :type data_test_csv: filepath (str)
    :param data_hdf5: If the dataset is in the hdf5 format, this is used instead
           of the csv file.
    :type data_hdf5: filepath (str)
    :param data_train_hdf5: If the training set is in the hdf5 format, this is
           used instead of the csv file.
    :type data_train_hdf5: filepath (str)
    :param data_validation_hdf5: If the validation set is in the hdf5 format,
           this is used instead of the csv file.
    :type data_validation_hdf5: filepath (str)
    :param data_test_hdf5: If the test set is in the hdf5 format, this is
           used instead of the csv file.
    :type data_test_hdf5: filepath (str)
    :param train_set_metadata_json: If the dataset is in hdf5 format, this is
           the associated json file containing metadata.
    :type train_set_metadata_json: filepath (str)
    :param experiment_name: The name for the experiment.
    :type experiment_name: Str
    :param model_name: Name of the model that is being used.
    :type model_name: Str
    :param model_load_path: If this is specified the loaded model will be used
           as initialization (useful for transfer learning).
    :type model_load_path: filepath (str)
    :param model_resume_path: Resumes training of the model from the path
           specified. The difference with model_load_path is that also training
           statistics like the current epoch and the loss and performance so
           far are also resumed effectively cotinuing a previously interrupted
           training process.
    :type model_resume_path: filepath (str)
    :param skip_save_progress_weights: Skips saving the weights at the end of
           each epoch. If this is true, training cannot be resumed from the
           exactly the state at the end of the previous epoch.
    :type skip_save_progress_weights: Boolean
    :param skip_save_processed_input: If a CSV dataset is provided it is
           preprocessed and then saved as an hdf5 and json to avoid running
           the preprocessing again. If this parameter is False,
           the hdf5 and json file are not saved.
    :type skip_save_processed_input: Boolean
    :param skip_save_unprocessed_output: By default predictions and
           their probabilities are saved in both raw unprocessed numpy files
           contaning tensors and as postprocessed CSV files
           (one for each output feature). If this parameter is True,
           only the CSV ones are saved and the numpy ones are skipped.
    :type skip_save_unprocessed_output: Boolean
    :param output_directory: The directory that will contanin the training
           statistics, the saved model and the training procgress files.
    :type output_directory: filepath (str)
    :param gpus: List of GPUs that are available for training.
    :type gpus: List
    :param gpu_fraction: Fraction of the memory of each GPU to use at
           the beginning of the training. The memory may grow elastically.
    :type gpu_fraction: Integer
    :param random_seed: Random seed used for weights initialization,
           splits and any other random function.
    :type random_seed: Integer
    :param debug: If true turns on tfdbg with inf_or_nan checks.
    :type debug: Boolean
    """
    # set input features defaults
    if model_definition_file is not None:
        with open(model_definition_file, 'r') as def_file:
            model_definition = merge_with_defaults(yaml.load(def_file))
    else:
        model_definition = merge_with_defaults(model_definition)

    # setup directories and file names
    experiment_dir_name = None
    if model_resume_path is not None:
        if os.path.exists(model_resume_path):
            experiment_dir_name = model_resume_path
        else:
            if is_on_master():
                logging.info(
                    'Model resume path does not exists, '
                    'starting training from scratch'
                )
            model_resume_path = None

    if model_resume_path is None:
        if is_on_master():
            experiment_dir_name = get_experiment_dir_name(
                output_directory,
                experiment_name,
                model_name
            )
        else:
            experiment_dir_name = '/'
    description_fn, training_stats_fn, model_dir = get_file_names(
        experiment_dir_name
    )

    # save description
    description = get_experiment_description(
        model_definition,
        data_csv,
        data_train_csv,
        data_validation_csv,
        data_test_csv,
        data_hdf5,
        data_train_hdf5,
        data_validation_hdf5,
        data_test_hdf5,
        train_set_metadata_json,
        random_seed
    )
    if is_on_master():
        save_json(description_fn, description)
        # print description
        logging.info('Experiment name: {}'.format(experiment_name))
        logging.info('Model name: {}'.format(model_name))
        logging.info('Output path: {}'.format(experiment_dir_name))
        logging.info('')
        for key, value in description.items():
            logging.info('{}: {}'.format(key, pformat(value, indent=4)))
        logging.info('')

    # preprocess
    (
        training_set,
        validation_set,
        test_set,
        train_set_metadata
    ) = preprocess_for_training(
        model_definition,
        data_csv=data_csv,
        data_train_csv=data_train_csv,
        data_validation_csv=data_validation_csv,
        data_test_csv=data_test_csv,
        data_hdf5=data_hdf5,
        data_train_hdf5=data_train_hdf5,
        data_validation_hdf5=data_validation_hdf5,
        data_test_hdf5=data_test_hdf5,
        train_set_metadata_json=train_set_metadata_json,
        skip_save_processed_input=skip_save_processed_input,
        preprocessing_params=model_definition[
            'preprocessing'],
        random_seed=random_seed
    )
    if is_on_master():
        logging.info('Training set: {0}'.format(training_set.size))
        logging.info('Validation set: {0}'.format(validation_set.size))
        logging.info('Test set: {0}'.format(test_set.size))

    # update model definition with metadata properties
    update_model_definition_with_metadata(model_definition, train_set_metadata)

    # run the experiment
    model, training_results = train(
        training_set=training_set,
        validation_set=validation_set,
        test_set=test_set,
        model_definition=model_definition,
        save_path=model_dir,
        model_load_path=model_load_path,
        resume=model_resume_path is not None,
        skip_save_progress_weights=skip_save_progress_weights,
        gpus=gpus,
        gpu_fraction=gpu_fraction,
        use_horovod=use_horovod,
        random_seed=random_seed,
        debug=debug
    )
    (
        train_trainset_stats,
        train_valisest_stats,
        train_testset_stats
    ) = training_results

    if is_on_master():
        # save train set metadata
        save_json(
            os.path.join(
                model_dir,
                TRAIN_SET_METADATA_FILE_NAME
            ),
            train_set_metadata
        )

    # grab the results of the model with highest validation test performance
    validation_field = model_definition['training']['validation_field']
    validation_measure = model_definition['training']['validation_measure']
    validation_field_result = train_valisest_stats[validation_field]

    # max or min depending on the measure
    best_function = get_best_function(validation_measure)
    epoch_best_vali_measure, best_vali_measure = best_function(
        enumerate(validation_field_result[validation_measure]),
        key=lambda pair: pair[1]
    )

    best_vali_measure_epoch_test_measure = train_testset_stats[
        validation_field
    ][validation_measure][epoch_best_vali_measure]

    # print the results of the model with highest validation test performance
    if is_on_master():
        logging.info('Best validation model epoch: {0}'.format(
            epoch_best_vali_measure + 1)
        )
        logging.info(
            'Best validation model {0} on validation set {1}: {2}'.format(
                validation_measure,
                validation_field,
                best_vali_measure)
        )
        logging.info('Best validation model {0} on test set {1}: {2}'.format(
            validation_measure,
            validation_field,
            best_vali_measure_epoch_test_measure)
        )

    # save training statistics
    if is_on_master():
        save_json(
            training_stats_fn,
            {
                'train': train_trainset_stats,
                'validation': train_valisest_stats,
                'test': train_testset_stats
            }
        )

    # predict
    test_results = predict(
        test_set,
        train_set_metadata,
        model,
        model_definition,
        model_definition['training']['batch_size'],
        only_predictions=False,
        gpus=gpus,
        gpu_fraction=gpu_fraction,
        debug=debug
    )
    model.close_session()

    # postprocess
    postprocessed_output = postprocess(
        test_results,
        model_definition['output_features'],
        train_set_metadata,
        experiment_dir_name,
        skip_save_unprocessed_output or not is_on_master()
    )

    if is_on_master():
        print_prediction_results(test_results)

        save_prediction_outputs(postprocessed_output, experiment_dir_name)
        save_prediction_statistics(test_results, experiment_dir_name)

        logging.info('\nFinished: {0}_{1}'.format(experiment_name, model_name))
        logging.info('Saved to: {}'.format(experiment_dir_name))

    return experiment_dir_name


def cli(sys_argv):
    parser = argparse.ArgumentParser(
        description='This script trains and tests a model.',
        prog='ludwig experiment',
        usage='%(prog)s [options]'
    )

    # ----------------------------
    # Experiment naming parameters
    # ----------------------------
    parser.add_argument(
        '--output_directory',
        type=str,
        default='results',
        help='directory that contains the results'
    )
    parser.add_argument(
        '--experiment_name',
        type=str,
        default='experiment',
        help='experiment name'
    )
    parser.add_argument(
        '--model_name',
        type=str,
        default='run',
        help='name for the model'
    )

    # ---------------
    # Data parameters
    # ---------------
    parser.add_argument(
        '--data_csv',
        help='input data CSV file. If it has a split column, it will be used '
             'for splitting (0: train, 1: validation, 2: test), otherwise the '
             'dataset will be randomly split'
    )
    parser.add_argument('--data_train_csv', help='input train data CSV file')
    parser.add_argument(
        '--data_validation_csv',
        help='input validation data CSV file'
    )
    parser.add_argument('--data_test_csv', help='input test data CSV file')

    parser.add_argument(
        '--data_hdf5',
        help='input data HDF5 file. It is an intermediate preprocess version of'
             ' the input CSV created the first time a CSV file is used in the '
             'same directory with the same name and a hdf5 extension'
    )
    parser.add_argument(
        '--data_train_hdf5',
        help='input train data HDF5 file. It is an intermediate preprocess '
             'version of the input CSV created the first time a CSV file is '
             'used in the same directory with the same name and a hdf5 '
             'extension'
    )
    parser.add_argument(
        '--data_validation_hdf5',
        help='input validation data HDF5 file. It is an intermediate preprocess'
             ' version of the input CSV created the first time a CSV file is '
             'used in the same directory with the same name and a hdf5 '
             'extension'
    )
    parser.add_argument(
        '--data_test_hdf5',
        help='input test data HDF5 file. It is an intermediate preprocess '
             'version of the input CSV created the first time a CSV file is '
             'used in the same directory with the same name and a hdf5 '
             'extension'
    )

    parser.add_argument(
        '--metadata_json',
        help='input metadata JSON file. It is an intermediate preprocess file'
             ' containing the mappings of the input CSV created the first time '
             'a CSV file is used in the same directory with the same name and a'
             ' json extension'
    )

    parser.add_argument(
        '-sspi',
        '--skip_save_processed_input',
        help='skips saving intermediate HDF5 and JSON files',
        action='store_true',
        default=False
    )
    parser.add_argument(
        '-ssuo',
        '--skip_save_unprocessed_output',
        help='skips saving intermediate NPY output files',
        action='store_true',
        default=False
    )

    # ----------------
    # Model parameters
    # ----------------
    model_definition = parser.add_mutually_exclusive_group(required=True)
    model_definition.add_argument(
        '-md',
        '--model_definition',
        type=yaml.load,
        help='model definition'
    )
    model_definition.add_argument(
        '-mdf',
        '--model_definition_file',
        help='YAML file describing the model. Ignores --model_hyperparameters'
    )

    parser.add_argument(
        '-mlp',
        '--model_load_path',
        help='path of a pretrained model to load as initialization'
    )
    parser.add_argument(
        '-mrp',
        '--model_resume_path',
        help='path of a the model directory to resume training of'
    )
    parser.add_argument(
        '-sspw',
        '--skip_save_progress_weights',
        help='does not save weights after each epoch. By default Ludwig saves '
             'weights after each epoch for enabling resuming of training, but '
             'if the model is really big that can be time consuming and will '
             'use twice as much storage space, use this parameter to skip it.'
    )

    # ------------------
    # Runtime parameters
    # ------------------
    parser.add_argument(
        '-rs',
        '--random_seed',
        type=int,
        default=42,
        help='a random seed that is going to be used anywhere there is a call '
             'to a random number generator: data splitting, parameter '
             'initialization and training set shuffling'
    )
    parser.add_argument(
        '-g',
        '--gpus',
        nargs='+',
        type=int,
        default=None,
        help='list of GPUs to use'
    )
    parser.add_argument(
        '-gf',
        '--gpu_fraction',
        type=float,
        default=1.0,
        help='fraction of gpu memory to initialize the process with'
    )
    parser.add_argument(
        '-uh',
        '--use_horovod',
        action='store_true',
        default=False,
        help='uses horovod for distributed training'
    )
    parser.add_argument(
        '-dbg',
        '--debug',
        action='store_true',
        default=False,
        help='enables debugging mode'
    )
    parser.add_argument(
        '-l',
        '--logging_level',
        default='info',
        help='the level of logging to use',
        choices=['critical', 'error', 'warning', 'info', 'debug', 'notset']
    )

    args = parser.parse_args(sys_argv)

    logging.basicConfig(
        stream=sys.stdout,
        level=logging_level_registry[args.logging_level],
        format='%(message)s'
    )

    set_on_master(args.use_horovod)

    if is_on_master():
        print_ludwig('Experiment', LUDWIG_VERSION)

    experiment(**vars(args))


if __name__ == '__main__':
    cli(sys.argv[1:])



# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import tensorflow as tf


def optimize(loss, training_parameters, learning_rate, global_step,
             horovod=None):
    if training_parameters is not None and training_parameters['decay'] == True:
        learning_rate = tf.train.exponential_decay(
            learning_rate, global_step,
            training_parameters['decay_steps'],
            training_parameters['decay_rate'],
            staircase=training_parameters['staircase'])

    with tf.variable_scope('optimizer'):
        if training_parameters is not None:
            optimizer_args = dict(training_parameters['optimizer'])
            optimizer_type = optimizer_args.pop('type')
            optimizer_fun = get_optimizer_fun(optimizer_type)
            if optimizer_type == 'adagradda':
                optimizer = optimizer_fun(learning_rate, global_step,
                                          **optimizer_args)
            else:
                optimizer = optimizer_fun(learning_rate, **optimizer_args)

            if horovod:
                optimizer = horovod.DistributedOptimizer(optimizer)

            optimize = optimizer.minimize(loss,
                                          global_step=global_step)
            if 'gradient_clipping' in training_parameters and \
                    training_parameters['gradient_clipping'] is not None:
                grad_clip_norm = training_parameters['gradient_clipping']
                gradients, variables = zip(*optimizer.compute_gradients(loss))
                gradients, _ = tf.clip_by_global_norm(gradients, grad_clip_norm)
                apply_grads = optimizer.apply_gradients(
                    zip(gradients, variables))
                increment_global_step = tf.assign(global_step, global_step + 1)
                optimize = tf.group(apply_grads, increment_global_step)
        else:
            optimizer = tf.train.AdamOptimizer(learning_rate)

            if horovod:
                optimizer = horovod.DistributedOptimizer(optimizer)

            optimize = optimizer.minimize(loss,
                                          global_step=global_step)

    return optimize, learning_rate


def get_optimizer_fun(optimizer_type):
    optimizer_type = optimizer_type.lower()
    if (
            optimizer_type == 'sgd' or
            optimizer_type == 'stochastic_gradient_descent' or
            optimizer_type == 'gd' or
            optimizer_type == 'gradient_descent'
    ):
        return tf.train.GradientDescentOptimizer
    elif optimizer_type == 'adam':
        return tf.train.AdamOptimizer
    elif optimizer_type == 'adadelta':
        return tf.train.AdadeltaOptimizer
    elif optimizer_type == 'adagrad':
        return tf.train.AdagradOptimizer
    elif optimizer_type == 'adagradda':
        return tf.train.AdagradDAOptimizer
    elif optimizer_type == 'momentum':
        return tf.train.MomentumOptimizer
    elif optimizer_type == 'ftrl':
        return tf.train.FtrlOptimizer
    elif optimizer_type == 'proximalgd':
        return tf.train.ProximalGradientDescentOptimizer
    elif optimizer_type == 'proximaladagrad':
        return tf.train.ProximalAdagradOptimizer
    elif optimizer_type == 'rmsprop':
        return tf.train.RMSPropOptimizer
    else:
        raise ValueError('Invalid optimizer_type: ' + optimizer_type)



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
BINARY = 'binary'
CATEGORY = 'category'
NUMERICAL = 'numerical'
SET = 'set'
BAG = 'bag'
TEXT = 'text'
SEQUENCE = 'sequence'
TIMESERIES = 'timeseries'
IMAGE = 'image'
HEIGHT = 'height'
WIDTH = 'width'
CROP_OR_PAD = 'crop_or_pad'
INTERPOLATE = 'interpolate'
LOSS = 'loss'
EVAL_LOSS = 'eval_loss'
TRAIN_MEAN_LOSS = 'train_mean_loss'
SOFTMAX_CROSS_ENTROPY = 'softmax_cross_entropy'
SAMPLED_SOFTMAX_CROSS_ENTROPY = 'sampled_softmax_cross_entropy'
ACCURACY = 'accuracy'
HITS_AT_K = 'hits_at_k'
MEAN_HITS_AT_K = 'mean_hits_at_k'
ERROR = 'error'
ABSOLUTE_ERROR = 'absolute_error'
SQUARED_ERROR = 'squared_error'
MEAN_SQUARED_ERROR = 'mean_squared_error'
MEAN_ABSOLUTE_ERROR = 'mean_absolute_error'
R2 = 'r2'
EDIT_DISTANCE = 'edit_distance'
PERPLEXITY = 'perplexity'
JACCARD = 'jaccard'
PREDICTIONS = 'predictions'
TOP_K_PREDICTIONS = 'top_k_predictions'
PROBABILITIES = 'probabilities'
CORRECT_PREDICTIONS = 'correct_predictions'
CORRECT_LAST_PREDICTIONS = 'correct_last_predictions'
CORRECT_OVERALL_PREDICTIONS = 'correct_overall_predictions'
CORRECT_ROWWISE_PREDICTIONS = 'correct_rowwise_predictions'
ROWWISE_ACCURACY = 'rowwise_accuracy'
LAST_ACCURACY = 'last_accuracy'
OVERALL_ACCURACY = 'overall_accuracy'
LAST_PROBABILTIES = 'last_probabilities'
LAST_PREDICTIONS = 'last_predictions'
LENGTHS = 'lengths'

FILL_WITH_CONST = 'fill_with_const'
FILL_WITH_MODE = 'fill_with_mode'
FILL_WITH_MEAN = 'fill_with_mean'
BACKFILL = 'backfill'

MEASURE = 'measure'
PREDICTION = 'prediction'

SUM = 'sum'
APPEND = 'append'
SEQ_SUM = 'seq_sum'
AVG_EXP = 'avg_exp'



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import json
import logging
import re
import unicodedata
from collections import Counter

import numpy as np

from ludwig.utils.misc import get_from_registry
from ludwig.utils.nlp_utils import load_nlp_pipeline, process_text

UNKNOWN_SYMBOL = '<UNK>'
PADDING_SYMBOL = '<PAD>'

SPLIT_REGEX = re.compile(r'\s+')
SPLIT_PUNCTUATION_REGEX = re.compile(r'\w+|[^\w\s]')
COMMA_REGEX = re.compile(r'\s*,\s*')
UNDERSCORE_REGEX = re.compile(r'\s*_\s*')


def make_safe_filename(s):
    def safe_char(c):
        if c.isalnum():
            return c
        else:
            return '_'

    return ''.join(safe_char(c) for c in s).rstrip('_')


def strip_accents(s):
    return ''.join(c for c in unicodedata.normalize('NFD', s)
                   if unicodedata.category(c) != 'Mn')


def str2bool(v):
    return v.lower() in ('yes', 'true', 't', '1')


def match_replace(string_to_match, list_regex):
    """Matches strings against regular expressions.

    arguments:
    string_to_match -- the string to match

    returns:
    string_to_match -- the cleaned string
    matched -- the list of regular expressions that matched
    """
    matched = []
    for regex in list_regex:
        match = re.search(regex[0], string_to_match)
        if match:
            string_to_match = re.sub(regex[0], regex[1], string_to_match)
            matched.append(regex[0].pattern)
    return string_to_match, matched


def create_vocabulary(data, format='space', custom_vocabulary=(),
                      add_unknown=True, add_padding=True,
                      lowercase=True,
                      num_most_frequent=None):
    max_line_length = 0
    unit_counts = Counter()

    if format == 'custom':
        vocab = sorted(list(set(custom_vocabulary)))
    else:
        format_function = get_from_registry(
            format,
            format_registry
        )
        for line in data:
            processed_line = format_function(
                line.lower() if lowercase else line)
            unit_counts.update(processed_line)
            max_line_length = max(max_line_length, len(processed_line))

        vocab = [unit for unit, count in
                 unit_counts.most_common(num_most_frequent)]

    if add_unknown:
        vocab = [UNKNOWN_SYMBOL] + vocab
    if add_padding:
        vocab = [PADDING_SYMBOL] + vocab

    str2idx = {unit: i for i, unit in enumerate(vocab)}
    str2freq = {unit: unit_counts.get(unit) if unit in unit_counts else 0 for
                unit in vocab}

    return vocab, str2idx, str2freq, max_line_length


def get_sequence_vector(sequence, format, unit_to_id, lowercase=True):
    format_function = get_from_registry(
        format,
        format_registry
    )
    format_dtype = get_from_registry(
        format,
        format_dtype_registry
    )
    return _get_sequence_vector(sequence, format_function, format_dtype,
                                unit_to_id, lowercase=lowercase)


def _get_sequence_vector(sequence, format_function, format_dtype, unit_to_id,
                         lowercase=True):
    unit_sequence = format_function(sequence.lower() if lowercase else sequence)
    unit_indices_vector = np.empty(len(unit_sequence), dtype=format_dtype)
    for i in range(len(unit_sequence)):
        curr_unit = unit_sequence[i]
        if curr_unit in unit_to_id:
            unit_indices_vector[i] = unit_to_id[curr_unit]
        else:
            unit_indices_vector[i] = unit_to_id[UNKNOWN_SYMBOL]
    return unit_indices_vector


def build_sequence_matrix(sequences, inverse_vocabulary, format, length_limit,
                          padding_symbol, padding='right',
                          lowercase=True):
    format_function = get_from_registry(
        format,
        format_registry
    )
    format_dtype = get_from_registry(
        format,
        format_dtype_registry
    )
    max_length = 0
    unit_vectors = []
    for sequence in sequences:
        unit_indices_vector = _get_sequence_vector(sequence, format_function,
                                                   format_dtype,
                                                   inverse_vocabulary,
                                                   lowercase=lowercase)
        unit_vectors.append(unit_indices_vector)
        if len(unit_indices_vector) > max_length:
            max_length = len(unit_indices_vector)

    if max_length < length_limit:
        logging.debug('max length of {0}: {1} < limit: {2}'.format(
            format, max_length, length_limit
        ))
    max_length = length_limit
    sequence_matrix = np.full((len(sequences), max_length),
                              inverse_vocabulary[padding_symbol],
                              dtype=format_dtype)
    for i, vector in enumerate(unit_vectors):
        limit = min(vector.shape[0], max_length)
        if padding == 'right':
            sequence_matrix[i, :limit] = vector[:limit]
        else:  # if padding == 'left
            sequence_matrix[i, max_length - limit:] = vector[:limit]
    return sequence_matrix


def ids_array_to_string(matrix, idx2str):
    texts = []
    for row in matrix:
        texts.append(
            ' '.join(map(lambda x: idx2str[x], [x for x in row if x > 0])))
    return texts


def json_string_to_list(s):
    s = str.replace(s, '\'', '\'')
    return json.loads(str.replace(s, '\'', '\''))


def space_string_to_list(s):
    return SPLIT_REGEX.split(s.strip())


def space_punctuation_string_to_list(s):
    return SPLIT_PUNCTUATION_REGEX.split(s.strip())

def underscore_string_to_list(s):
    return UNDERSCORE_REGEX.split(s.strip())


def comma_string_to_list(s):
    return COMMA_REGEX.split(s.strip())


def untokenized_string_to_list(s):
    return [s]


def stripped_string_to_list(s):
    return [s.strip()]


def english_tokenize(text):
    return process_text(text, load_nlp_pipeline())


def english_tokenize_filter(text):
    return process_text(text, load_nlp_pipeline(),
                        filter_numbers=True,
                        filter_punctuation=True,
                        filter_short_tokens=True)


def english_tokenize_remove_stopwords(text):
    return process_text(text, load_nlp_pipeline(), filter_stopwords=True)


def english_lemmatize(text):
    return process_text(text, load_nlp_pipeline(), return_lemma=True)


def english_lemmatize_filter(text):
    return process_text(text, load_nlp_pipeline(),
                        return_lemma=True,
                        filter_numbers=True,
                        filter_punctuation=True,
                        filter_short_tokens=True)


def english_lemmatize_remove_stopwords(text):
    return process_text(text, load_nlp_pipeline(),
                        return_lemma=True,
                        filter_stopwords=True)


def characters_to_list(s):
    return s


format_registry = {
    'json': json_string_to_list,
    'space': space_string_to_list,
    'space_punct': space_punctuation_string_to_list,
    'underscore': underscore_string_to_list,
    'comma': comma_string_to_list,
    'untokenized': untokenized_string_to_list,
    'stripped': stripped_string_to_list,
    'english_tokenize': english_tokenize,
    'english_tokenize_filter': english_tokenize_filter,
    'english_tokenize_remove_stopwords': english_tokenize_remove_stopwords,
    'english_lemmatize': english_lemmatize,
    'english_lemmatize_filter': english_lemmatize_filter,
    'english_lemmatize_remove_stopwords': english_lemmatize_remove_stopwords,
    'characters': characters_to_list
}

format_dtype_registry = {
    'json': np.int32,
    'space': np.int32,
    'space_punct': np.int32,
    'underscore': np.int32,
    'comma': np.int32,
    'untokenized': np.int32,
    'stripped': np.int32,
    'english_tokenize': np.int32,
    'english_tokenize_filter': np.int32,
    'english_tokenize_remove_stopwords': np.int32,
    'english_lemmatize': np.int32,
    'english_lemmatize_filter': np.int32,
    'english_lemmatize_remove_stopwords': np.int32,
    'characters': np.int8
}



# coding=utf-8
'''Ludwig: a deep learning experimentation toolbox
'''
from codecs import open
from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

# Get the long description from the README.md file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='ludwig',

    version='0.1.0',

    description='A deep learning experimentation toolbox',
    long_description=long_description,
    long_description_content_type='text/markdown',

    url='https://ludwig.ai',

    author='Piero Molino',
    author_email='piero.molino@gmail.com',

    license='Apache 2.0',

    keywords='ludwig deep learning deep_learning machine machine_learning natural language processing computer vision',

    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    python_requires='>=3',

    include_package_data=True,
    package_data={'ludwig': ['etc/*', 'examples/*.py']},

    install_requires=['Cython>=0.25',
                      'h5py>=2.6',
                      'matplotlib>=2.0',
                      'numpy>=1.12',
                      'pandas>=0.19',
                      'scipy>=0.18',
                      'scikit-learn',
                      'scikit-image',
                      'seaborn>=0.7',
                      'spacy>=2.0',
                      'tqdm',
                      'tabulate>=0.7',
                      'tensorflow>=1.12',
                      'PyYAML>=3.12'
                      ],

    entry_points={
        'console_scripts': [
            'ludwig=ludwig.cli:main'
        ]
    }
)



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
import os
from collections import OrderedDict

import numpy as np
import tensorflow as tf

from ludwig.constants import *
from ludwig.features.base_feature import BaseFeature
from ludwig.features.base_feature import InputFeature
from ludwig.features.base_feature import OutputFeature
from ludwig.features.feature_utils import set_str_to_idx
from ludwig.models.modules.embedding_modules import EmbedSparse
from ludwig.models.modules.initializer_modules import get_initializer
from ludwig.utils.misc import set_default_value
from ludwig.utils.strings_utils import create_vocabulary


class SetBaseFeature(BaseFeature):
    def __init__(self, feature):
        super().__init__(feature)
        self.type = IMAGE

    preprocessing_defaults = {
        'format': 'space',
        'most_common': 10000,
        'lowercase': False,
        'missing_value_strategy': FILL_WITH_CONST,
        'fill_value': ''
    }

    @staticmethod
    def get_feature_meta(column, preprocessing_parameters):
        idx2str, str2idx, str2freq, max_size = create_vocabulary(
            column,
            preprocessing_parameters['format'],
            num_most_frequent=preprocessing_parameters['most_common'],
            lowercase=preprocessing_parameters['lowercase']
        )
        return {
            'idx2str': idx2str,
            'str2idx': str2idx,
            'str2freq': str2freq,
            'vocab_size': len(str2idx),
            'max_set_size': max_size
        }

    @staticmethod
    def feature_data(column, metadata, preprocessing_parameters):
        feature_vector = np.array(
            column.map(
                lambda x: set_str_to_idx(
                    x,
                    metadata['str2idx'],
                    preprocessing_parameters['format']
                )
            )
        )

        set_matrix = np.zeros(
            (len(column),
             len(metadata['str2idx'])),
            dtype=bool
        )

        for i in range(len(column)):
            set_matrix[i, feature_vector[i]] = 1

        return set_matrix

    @staticmethod
    def add_feature_data(
            feature,
            dataset_df,
            data,
            metadata,
            preprocessing_parameters,
    ):
        data[feature['name']] = SetBaseFeature.feature_data(
            dataset_df[feature['name']].astype(str),
            metadata[feature['name']],
            preprocessing_parameters
        )


class SetInputFeature(SetBaseFeature, InputFeature):
    def __init__(self, feature):
        super().__init__(feature)

        self.vocab = []
        self.embedding_size = 50
        self.representation = 'dense'
        self.embeddings_trainable = True
        self.pretrained_embeddings = None
        self.embeddings_on_cpu = False
        self.dropout = False
        self.initializer = None
        self.regularize = True

        _ = self.overwrite_defaults(feature)

        self.embed_sparse = EmbedSparse(
            self.vocab,
            self.embedding_size,
            representation=self.representation,
            embeddings_trainable=self.embeddings_trainable,
            pretrained_embeddings=self.pretrained_embeddings,
            embeddings_on_cpu=self.embeddings_on_cpu,
            dropout=self.dropout,
            initializer=self.initializer,
            regularize=self.regularize
        )

    def _get_input_placeholder(self):
        # None is for dealing with variable batch size
        return tf.placeholder(
            tf.int32,
            shape=[None, len(self.vocab)],
            name=self.name
        )

    def build_input(
            self,
            regularizer,
            dropout_rate,
            is_training=False,
            **kwargs
    ):
        placeholder = self._get_input_placeholder()
        logging.debug('  placeholder: {0}'.format(placeholder))

        embedded, embedding_size = self.embed_sparse(
            placeholder,
            regularizer,
            dropout_rate,
            is_training=is_training
        )
        logging.debug('  feature_representation: {0}'.format(embedded))

        feature_representation = {
            'name': self.name,
            'type': self.type,
            'representation': embedded,
            'size': embedding_size,
            'placeholder': placeholder
        }

        return feature_representation

    @staticmethod
    def update_model_definition_with_metadata(
            input_feature,
            feature_metadata,
            *args,
            **kwargs
    ):
        input_feature['vocab'] = feature_metadata['idx2str']

    @staticmethod
    def populate_defaults(input_feature):
        set_default_value(input_feature, 'tied_weights', None)


class SetOutputFeature(SetBaseFeature, OutputFeature):
    def __init__(self, feature):
        super().__init__(feature)
        self.type = SET

        self.loss = {'type': 'sigmoid_cross_entropy'}
        self.num_classes = 0
        self.threshold = 0.5
        self.initializer = None
        self.regularize = True

        _ = self.overwrite_defaults(feature)

    def _get_output_placeholder(self):
        return tf.placeholder(
            tf.bool,
            shape=[None, self.num_classes],
            name='{}_placeholder'.format(self.name)
        )

    def _get_predictions(
            self,
            hidden,
            hidden_size,
            regularizer=None
    ):
        if not self.regularize:
            regularizer = None

        with tf.variable_scope('predictions_{}'.format(self.name)):
            initializer_obj = get_initializer(self.initializer)
            weights = tf.get_variable(
                'weights',
                initializer=initializer_obj([hidden_size, self.num_classes]),
                regularizer=regularizer
            )
            logging.debug('  class_weights: {0}'.format(weights))

            biases = tf.get_variable(
                'biases',
                [self.num_classes]
            )
            logging.debug('  class_biases: {0}'.format(biases))

            logits = tf.matmul(hidden, weights) + biases
            logging.debug('  logits: {0}'.format(logits))

            probabilities = tf.nn.sigmoid(
                logits,
                name='probabilities_{}'.format(self.name)
            )

            predictions = tf.greater_equal(
                probabilities,
                self.threshold,
                name='predictions_{}'.format(self.name)
            )

        return predictions, probabilities, logits

    def _get_loss(
            self,
            targets,
            logits
    ):
        with tf.variable_scope('loss_{}'.format(self.name)):
            train_loss = tf.nn.sigmoid_cross_entropy_with_logits(
                labels=tf.to_float(targets),
                logits=logits
            )
            train_loss = tf.reduce_sum(train_loss, axis=1)

            train_mean_loss = tf.reduce_mean(
                train_loss,
                name='train_mean_loss_{}'.format(self.name)
            )

        return train_mean_loss, train_loss

    def _get_measures(self, targets, predictions):
        intersection = tf.reduce_sum(
            tf.to_float(tf.logical_and(targets, predictions)),
            axis=1
        )
        union = tf.reduce_sum(
            tf.to_float(tf.logical_or(targets, predictions)),
            axis=1
        )
        jaccard_index = intersection / union

        return jaccard_index

    def build_output(
            self,
            hidden,
            hidden_size,
            regularizer=None,
            **kwargs
    ):
        output_tensors = {}

        # ================ Placeholder ================
        targets = self._get_output_placeholder()
        output_tensors[self.name] = targets
        logging.debug('  targets_placeholder: {0}'.format(targets))

        # ================ Predictions ================
        ppl = self._get_predictions(
            hidden,
            hidden_size,
            regularizer=regularizer
        )
        predictions, probabilities, logits = ppl

        jaccard_index = self._get_measures(targets, predictions)

        output_tensors[PREDICTIONS + '_' + self.name] = predictions
        output_tensors[PROBABILITIES + '_' + self.name] = probabilities
        output_tensors[JACCARD + '_' + self.name] = jaccard_index

        # ================ Loss (Binary Cross Entropy) ================
        train_mean_loss, eval_loss = self._get_loss(targets, logits)

        output_tensors[EVAL_LOSS + '_' + self.name] = eval_loss
        output_tensors[TRAIN_MEAN_LOSS + '_' + self.name] = train_mean_loss

        tf.summary.scalar(
            'train_mean_loss_{}'.format(self.name),
            train_mean_loss
        )

        return train_mean_loss, eval_loss, output_tensors

    default_validation_measure = JACCARD

    output_config = OrderedDict([
        (LOSS, {
            'output': EVAL_LOSS,
            'aggregation': SUM,
            'value': 0,
            'type': MEASURE
        }),
        (JACCARD, {
            'output': JACCARD,
            'aggregation': SUM,
            'value': 0,
            'type': MEASURE
        }),
        (PREDICTIONS, {
            'output': PREDICTIONS,
            'aggregation': APPEND,
            'value': [],
            'type': PREDICTION
        }),
        (PROBABILITIES, {
            'output': PROBABILITIES,
            'aggregation': APPEND,
            'value': [],
            'type': PREDICTION
        })
    ])

    @staticmethod
    def update_model_definition_with_metadata(
            output_feature,
            feature_metadata,
            *args,
            **kwargs
    ):
        output_feature[LOSS]['type'] = None
        output_feature['num_classes'] = feature_metadata['vocab_size']

    @staticmethod
    def calculate_overall_stats(
            test_stats,
            output_feature,
            dataset,
            train_set_metadata
    ):
        pass

    @staticmethod
    def postprocess_results(
            output_feature,
            result,
            metadata,
            experiment_dir_name,
            skip_save_unprocessed_output=False
    ):
        postprocessed = {}
        npy_filename = os.path.join(experiment_dir_name, '{}_{}.npy')
        name = output_feature['name']

        if PREDICTIONS in result and len(result[PREDICTIONS]) > 0:
            preds = result[PREDICTIONS]
            if 'idx2str' in metadata:
                postprocessed[PREDICTIONS] = [
                    [metadata['idx2str'][i] for i, pred in enumerate(pred_set)
                     if pred == True] for pred_set in preds
                ]
            else:
                postprocessed[PREDICTIONS] = preds

            if not skip_save_unprocessed_output:
                np.save(npy_filename.format(name, PREDICTIONS), preds)

            del result[PREDICTIONS]

        if PROBABILITIES in result and len(result[PROBABILITIES]) > 0:
            probs = result[PROBABILITIES]
            prob = [[prob for prob in prob_set if
                     prob >= output_feature['threshold']] for prob_set in probs]
            postprocessed[PROBABILITIES] = probs
            postprocessed['probability'] = prob

            if not skip_save_unprocessed_output:
                np.save(npy_filename.format(name, PROBABILITIES), probs)
                np.save(npy_filename.format(name, 'probability'), probs)

            del result[PROBABILITIES]

        return postprocessed

    @staticmethod
    def populate_defaults(output_feature):
        set_default_value(output_feature, LOSS, {'weight': 1, 'type': None})
        set_default_value(output_feature[LOSS], 'weight', 1)

        set_default_value(output_feature, 'threshold', 0.5)
        set_default_value(output_feature, 'dependencies', [])
        set_default_value(output_feature, 'reduce_input', SUM)
        set_default_value(output_feature, 'reduce_dependencies', SUM)



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import copy
import os
import random
import subprocess
import sys
from collections import OrderedDict, Mapping

import numpy

import ludwig.globals


def get_experiment_description(model_definition,
                               dataset_type='generic',
                               data_csv=None,
                               data_hdf5=None,
                               metadata_json=None,
                               data_train_csv=None,
                               data_validation_csv=None,
                               data_test_csv=None,
                               data_train_hdf5=None,
                               data_validation_hdf5=None,
                               data_test_hdf5=None,
                               random_seed=None):
    is_a_git_repo = subprocess.call(['git', 'branch'],
                                    stderr=subprocess.STDOUT,
                                    stdout=open(os.devnull, 'w')) == 0

    description = OrderedDict()
    description['ludwig_version'] = ludwig.globals.LUDWIG_VERSION
    description['command'] = ' '.join(sys.argv)
    if is_a_git_repo:
        description['commit_hash'] = \
            subprocess.check_output(['git', 'rev-parse', 'HEAD']).decode(
                'utf-8')[:12]
    description['dataset_type'] = dataset_type
    if random_seed is not None:
        description['random_seed'] = random_seed
    if data_csv is not None:
        description['input_data'] = data_csv
    elif data_hdf5 is not None and metadata_json is not None:
        description['input_data'] = data_hdf5
        description['input_metadata'] = metadata_json
    elif data_train_csv is not None:
        description['input_data_train'] = data_train_csv
        if data_validation_csv is not None:
            description['input_data_validation'] = data_validation_csv
        if data_test_csv is not None:
            description['input_data_test'] = data_test_csv
    elif data_train_hdf5 is not None and metadata_json is not None:
        description['input_data_train'] = data_train_hdf5
        if data_validation_hdf5 is not None:
            description['input_data_validation'] = data_validation_hdf5
        if data_test_hdf5 is not None:
            description['input_data_test'] = data_test_hdf5
        description['input_metadata'] = metadata_json
    description['model_definition'] = model_definition

    return description


def set_random_seed(random_seed):
    os.environ['PYTHONHASHSEED'] = str(random_seed)
    random.seed(random_seed)
    numpy.random.seed(random_seed)


def merge_dict(dct, merge_dct):
    """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.
    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None
    """
    dct = copy.deepcopy(dct)
    for k, v in merge_dct.items():
        if (k in dct and isinstance(dct[k], dict)
                and isinstance(merge_dct[k], Mapping)):
            dct[k] = merge_dict(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]
    return dct


def sum_dicts(dicts, dict_type=dict):
    summed_dict = dict_type()
    for d in dicts:
        for key, value in d.items():
            if key in summed_dict:
                prev_value = summed_dict[key]
                if isinstance(value, (dict, OrderedDict)):
                    summed_dict[key] = sum_dicts([prev_value, value],
                                                 dict_type=type(value))
                elif isinstance(value, numpy.ndarray):
                    summed_dict[key] = numpy.concatenate((prev_value, value))
                else:
                    summed_dict[key] = prev_value + value
            else:
                summed_dict[key] = value
    return summed_dict


def get_from_registry(key, registry):
    if hasattr(key, 'lower'):
        key = key.lower()
    if key in registry:
        return registry[key]
    else:
        raise ValueError(
            'Key {} not supported, available options: {}'.format(
                key, registry.keys()
            )
        )


def set_default_value(dictionary, key, value):
    if key not in dictionary:
        dictionary[key] = value



# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from collections import OrderedDict

import tensorflow as tf

from ludwig.features.feature_registries import input_type_registry
from ludwig.utils.algorithms_utils import topological_sort_feature_dependencies
from ludwig.utils.misc import get_from_registry


def build_inputs(input_features,
                 regularizer,
                 dropout_rate,
                 is_training=True,
                 **kwargs):
    # ================ Inputs =============
    feature_representations = OrderedDict()
    input_features = topological_sort_feature_dependencies(input_features)
    for input_feature in input_features:
        feature_representation = build_single_input(input_feature,
                                                    regularizer,
                                                    dropout_rate,
                                                    is_training=is_training,
                                                    **kwargs)
        feature_representations[input_feature['name']] = feature_representation
    return feature_representations


def build_single_input(input_feature,
                       regularizer,
                       dropout_rate,
                       is_training=True,
                       **kwargs):
    scope_name = input_feature['name']
    logging.debug('- Input {} feature {}'.format(
        input_feature['type'],
        input_feature['name']
    ))
    if input_feature['tied_weights'] is not None:
        scope_name = input_feature['tied_weights']

    with tf.variable_scope(scope_name, reuse=tf.AUTO_REUSE):
        input_feature_class = get_from_registry(
            input_feature['type'],
            input_type_registry
        )
        input_feature_obj = input_feature_class(input_feature)
        feature_representation = input_feature_obj.build_input(
            regularizer=regularizer,
            dropout_rate=dropout_rate, is_training=is_training,
            **kwargs)
        feature_representation['representation'] = tf.identity(
            feature_representation['representation'],
            name=scope_name)
    return feature_representation


dynamic_length_encoders = {
    'rnn',
    'embed'
}



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import pandas as pd

from ludwig.features.feature_registries import output_type_registry
from ludwig.features.feature_utils import SEQUENCE_TYPES
from ludwig.utils.misc import get_from_registry


def postprocess_results(
        result,
        output_feature,
        metadata,
        experiment_dir_name='',
        skip_save_unprocessed_output=False
):
    feature = get_from_registry(
        output_feature['type'], output_type_registry
    )
    return feature.postprocess_results(
        output_feature,
        result,
        metadata,
        experiment_dir_name,
        skip_save_unprocessed_output=skip_save_unprocessed_output
    )


def postprocess(
        results,
        output_features,
        metadata,
        experiment_dir_name='',
        skip_save_unprocessed_output=False
):
    postprocessed = {}
    for output_feature in output_features:
        postprocessed[output_feature['name']] = postprocess_results(
            results[output_feature['name']],
            output_feature,
            metadata.get(output_feature['name'], {}),
            experiment_dir_name=experiment_dir_name,
            skip_save_unprocessed_output=skip_save_unprocessed_output
        )
    return postprocessed


def postprocess_df(model_output, output_features, metadata):
    postprocessed_output = postprocess(
        model_output,
        output_features,
        metadata,
        skip_save_unprocessed_output=True
    )
    data_for_df = {}
    for output_feature in output_features:
        output_feature_name = output_feature['name']
        output_feature_type = output_feature['type']
        output_feature_dict = postprocessed_output[output_feature_name]
        for key_val in output_feature_dict.items():
            output_subgroup_name, output_type_value = key_val
            if (hasattr(output_type_value, 'shape') and
                len(output_type_value.shape)) > 1:
                if output_feature_type in SEQUENCE_TYPES:
                    data_for_df[
                        '{}_{}'.format(
                            output_feature_name,
                            output_subgroup_name
                        )
                    ] = output_type_value.tolist()
                else:
                    for i, value in enumerate(output_type_value.T):
                        if (output_feature_name in metadata and
                                'idx2str' in metadata[output_feature_name]):
                            class_name = metadata[output_feature_name][
                                'idx2str'][i]
                        else:
                            class_name = str(i)
                        data_for_df[
                            '{}_{}_{}'.format(
                                output_feature_name,
                                output_subgroup_name,
                                class_name
                            )
                        ] = value
            else:
                data_for_df[
                    '{}_{}'.format(
                        output_feature_name,
                        output_subgroup_name
                    )
                ] = output_type_value
    output_df = pd.DataFrame(data_for_df)
    return output_df



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
import re
from html.parser import HTMLParser

from ludwig.utils import strings_utils


class HTMLStripper(HTMLParser):
    def __init__(self):
        super().__init__()
        self.reset()
        self.strict = False
        self.convert_charrefs = True
        self.fed = []

    def handle_data(self, data):
        self.fed.append(data)

    def get_data(self):
        return ''.join(self.fed)

    def error(self, message):
        logging.error(message)
        pass


def strip_tags(html):
    stripper = HTMLStripper()
    stripper.feed(html)
    return stripper.get_data()


# regular expressions for cleaning text
res_pre = [
    (re.compile(r'([^.:;\?\!>])(<br/?>)'), r'\1.\2'),
    (re.compile(r'<br/?>'), r' ')
]
res_post = [
    (re.compile(r'[ \t\0]'), r' '),
    (re.compile(r'[–_]'), r'-'),
    (re.compile(r'[\’\‘]'), r'''),
    (re.compile(r'[”“]]'), r'''),
    (re.compile(r'℅'), r'%'),
    (re.compile(r'([^.>])(<br/?>)'), r'\1.\2'),
    (re.compile(r'\\\\[NnRr]'), r' '),
    (re.compile(r'\\[NnRr]'), r' '),
    (re.compile(r'[\n\r]'), r' '),
    (re.compile(r'\\\\'), r' / '),
    (re.compile(r'<br/?>'), r' '),
    (re.compile(r'\\\\'''), r'\''),
    (re.compile(r'^\'([^\']+)$'), r'\1'),
    (re.compile(r'([\<\>\{\}\[\]\(\)\-\+\=:;,\./\?\!\$%&£#@\'₹ ])\1+'), r'\1'),
    (re.compile(
        r'[^qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM1234567890\<\>\{\}\[\]\(\)\-\+\=:;,\./\?\!\$%&£#@\'₹ ]'),
     r' '),
    (re.compile(r'\s{2,}'), r' ')
]


def clean_html(html_text):
    # print()
    # print(html_text)
    html_text, matched = strings_utils.match_replace(html_text, res_pre)
    # print(html_text)
    html_text = strip_tags(html_text)
    # print(html_text)
    html_text = strings_utils.strip_accents(html_text)
    # print(html_text)
    # result = html_text.strip(
    #     'qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM1234567890\<\>\{\}\[\]\(\)\-\+\=:;,\./\?\!\$%&€£#@'₹\' ')
    # if result:
    #     print(result)
    html_text, matched = strings_utils.match_replace(html_text, res_post)
    # print(matched)
    # print(html_text)
    return html_text



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
import os
from collections import OrderedDict

import numpy as np
import tensorflow as tf

from ludwig.constants import *
from ludwig.features.base_feature import BaseFeature
from ludwig.features.base_feature import InputFeature
from ludwig.features.base_feature import OutputFeature
from ludwig.models.modules.embedding_modules import Embed
from ludwig.models.modules.initializer_modules import get_initializer
from ludwig.models.modules.loss_modules import mean_confidence_penalty
from ludwig.models.modules.loss_modules import sampled_softmax_cross_entropy
from ludwig.models.modules.loss_modules import weighted_softmax_cross_entropy
from ludwig.models.modules.measure_modules import accuracy as get_accuracy
from ludwig.models.modules.measure_modules import hits_at_k as get_hits_at_k
from ludwig.utils.math_utils import int_type
from ludwig.utils.math_utils import softmax
from ludwig.utils.metrics_utils import ConfusionMatrix
from ludwig.utils.misc import set_default_value
from ludwig.utils.strings_utils import UNKNOWN_SYMBOL
from ludwig.utils.strings_utils import create_vocabulary


class CategoryBaseFeature(BaseFeature):
    def __init__(self, feature):
        super().__init__(feature)
        self.type = CATEGORY

    preprocessing_defaults = {
        'most_common': 10000,
        'lowercase': False,
        'missing_value_strategy': FILL_WITH_CONST,
        'fill_value': UNKNOWN_SYMBOL
    }

    @staticmethod
    def get_feature_meta(column, preprocessing_parameters):
        idx2str, str2idx, str2freq, _ = create_vocabulary(
            column, 'stripped',
            num_most_frequent=preprocessing_parameters['most_common'],
            lowercase=preprocessing_parameters['lowercase'],
            add_padding=False
        )
        return {
            'idx2str': idx2str,
            'str2idx': str2idx,
            'str2freq': str2freq,
            'vocab_size': len(str2idx)
        }

    @staticmethod
    def feature_data(column, metadata):
        return np.array(
            column.map(
                lambda x: (
                    metadata['str2idx'][x] if x in metadata['str2idx']
                    else metadata['str2idx'][UNKNOWN_SYMBOL]
                )
            ),
            dtype=int_type(metadata['vocab_size'])
        )

    @staticmethod
    def add_feature_data(
            feature,
            dataset_df,
            data,
            metadata,
            preprocessing_parameters=None
    ):
        data[feature['name']] = CategoryBaseFeature.feature_data(
            dataset_df[feature['name']].astype(str),
            metadata[feature['name']]
        )


class CategoryInputFeature(CategoryBaseFeature, InputFeature):
    def __init__(self, feature):
        super().__init__(feature)

        self.vocab = []

        self.embedding_size = 50
        self.representation = 'dense'
        self.embeddings_trainable = True
        self.pretrained_embeddings = None
        self.embeddings_on_cpu = False
        self.dropout = False
        self.initializer = None
        self.regularize = True

        _ = self.overwrite_defaults(feature)

        self.embed = Embed(
            vocab=self.vocab,
            embedding_size=self.embedding_size,
            representation=self.representation,
            embeddings_trainable=self.embeddings_trainable,
            pretrained_embeddings=self.pretrained_embeddings,
            embeddings_on_cpu=self.embeddings_on_cpu,
            dropout=self.dropout,
            initializer=self.initializer,
            regularize=self.regularize
        )

    @staticmethod
    def update_model_definition_with_metadata(
            input_feature,
            feature_metadata,
            *args,
            **kwargs
    ):
        input_feature['vocab'] = feature_metadata['idx2str']

    def _get_input_placeholder(self):
        return tf.placeholder(
            tf.int32,
            shape=[None],  # None is for dealing with variable batch size
            name='{}_placeholder'.format(self.name)
        )

    def build_input(
            self,
            regularizer,
            dropout_rate,
            is_training=False,
            **kwargs
    ):
        placeholder = self._get_input_placeholder()
        logging.debug('  placeholder: {0}'.format(placeholder))

        # ================ Embeddings ================
        embedded, embedding_size = self.embed(
            placeholder,
            regularizer,
            dropout_rate,
            is_training=is_training
        )
        logging.debug('  feature_representation: {0}'.format(
            embedded))

        feature_representation = {
            'name': self.name,
            'type': self.type,
            'representation': embedded,
            'size': embedding_size,
            'placeholder': placeholder
        }
        return feature_representation

    @staticmethod
    def populate_defaults(input_feature):
        set_default_value(input_feature, 'tied_weights', None)


class CategoryOutputFeature(CategoryBaseFeature, OutputFeature):
    def __init__(self, feature):
        super().__init__(feature)

        self.loss = {'type': SOFTMAX_CROSS_ENTROPY}
        self.num_classes = 0
        self.top_k = 3
        self.initializer = None
        self.regularize = True

        _ = self.overwrite_defaults(feature)

    default_validation_measure = ACCURACY

    output_config = OrderedDict([
        (LOSS, {
            'output': EVAL_LOSS,
            'aggregation': SUM,
            'value': 0,
            'type': MEASURE
        }),
        (ACCURACY, {
            'output': CORRECT_PREDICTIONS,
            'aggregation': SUM,
            'value': 0,
            'type': MEASURE
        }),
        (HITS_AT_K, {
            'output': HITS_AT_K,
            'aggregation': SUM,
            'value': 0,
            'type': MEASURE
        }),
        (PREDICTIONS, {
            'output': PREDICTIONS,
            'aggregation': APPEND,
            'value': [],
            'type': PREDICTION
        }),
        (PROBABILITIES, {
            'output': PROBABILITIES,
            'aggregation': APPEND,
            'value': [],
            'type': PREDICTION
        })
    ])

    def _get_output_placeholder(self):
        return tf.placeholder(
            tf.int64,
            [None],  # None is for dealing with variable batch size
            name='{}_placeholder'.format(self.name)
        )

    def _get_predictions(
            self,
            hidden,
            hidden_size,
            regularizer=None
    ):
        if not self.regularize:
            regularizer = None

        with tf.variable_scope('predictions_{}'.format(self.name)):
            initializer_obj = get_initializer(self.initializer)
            weights = tf.get_variable(
                'weights',
                initializer=initializer_obj([hidden_size, self.num_classes]),
                regularizer=regularizer
            )
            logging.debug('  class_weights: {0}'.format(weights))

            biases = tf.get_variable(
                'biases',
                [self.num_classes]
            )
            logging.debug('  class_biases: {0}'.format(biases))

            logits = tf.matmul(hidden, weights) + biases
            logging.debug('  logits: {0}'.format(logits))

            probabilities = tf.nn.softmax(
                logits,
                name='probabilities_{}'.format(self.name)
            )
            predictions = tf.argmax(
                logits,
                -1,
                name='predictions_{}'.format(self.name)
            )

            with tf.device('/cpu:0'):
                top_k_predictions = tf.nn.top_k(
                    logits,
                    k=self.top_k,
                    sorted=True,
                    name='top_k_predictions_{}'.format(self.name)
                )

        return (
            predictions,
            top_k_predictions,
            probabilities,
            logits,
            weights,
            biases
        )

    def _get_loss(
            self,
            targets,
            hidden,
            logits,
            probabilities,
            class_weights,
            class_biases
    ):
        with tf.variable_scope('loss_{}'.format(self.name)):
            if ('distances' in self.loss and
                    self.loss['distances'] is not None):

                class_distances = self.loss['distances']

                if (class_distances.shape[0] != self.num_classes or
                        class_distances.shape[1] != self.num_classes):
                    logging.info(
                        'Class distances is {} while num classes is {}'.format(
                            class_distances.shape,
                            self.num_classes
                        )
                    )
                    if (class_distances.shape[0] > self.num_classes and
                            class_distances.shape[1] > self.num_classes):
                        # keep only the first num_classes rows and columns
                        class_distances = class_distances[
                                          :self.num_classes,
                                          :self.num_classes
                                          ]
                    elif (class_distances.shape[0] < self.num_classes and
                          class_distances.shape[1] < self.num_classes):
                        # fill the missing parts of the matrix with 0s and 1
                        # on the diagonal
                        diag = np.diag((self.num_classes, self.num_classes))
                        diag[
                        :class_distances.shape[0],
                        :class_distances.shape[1]
                        ] = class_distances
                        class_distances = diag

                class_distances = tf.constant(
                    class_distances,
                    name='class_distances_{}'.format(self.name)
                )
                vector_labels = tf.gather(
                    class_distances,
                    targets,
                    name='vector_labels_{}'.format(self.name)
                )
            else:
                vector_labels = tf.one_hot(
                    targets,
                    self.num_classes,
                    name='vector_labels_{}'.format(self.name)
                )

            if self.loss['type'] == 'sampled_softmax_cross_entropy':
                train_loss, eval_loss = sampled_softmax_cross_entropy(
                    targets,
                    hidden,
                    logits,
                    vector_labels,
                    class_weights,
                    class_biases,
                    self.loss,
                    self.num_classes
                )
            elif self.loss['type'] == 'softmax_cross_entropy':
                train_loss = weighted_softmax_cross_entropy(
                    logits,
                    vector_labels,
                    self.loss
                )
                eval_loss = train_loss
            else:
                train_mean_loss = None
                eval_loss = None
                raise ValueError(
                    'Unsupported loss type {}'.format(self.loss['type'])
                )

            if self.loss['robust_lambda'] > 0:
                train_loss = ((1 - self.loss['robust_lambda']) * train_loss +
                              self.loss['robust_lambda'] / self.num_classes)

            train_mean_loss = tf.reduce_mean(
                train_loss,
                name='train_mean_loss_{}'.format(self.name)
            )

            if self.loss['confidence_penalty'] > 0:
                mean_penalty = mean_confidence_penalty(
                    probabilities,
                    self.num_classes
                )
                train_mean_loss += (
                        self.loss['confidence_penalty'] * mean_penalty
                )

        return train_mean_loss, eval_loss

    def _get_measures(self, targets, predictions, logits):
        with tf.variable_scope('measures_{}'.format(self.name)):
            accuracy_val, correct_predictions = get_accuracy(
                targets,
                predictions,
                self.name
            )
            hits_at_k_val, mean_hits_at_k = get_hits_at_k(
                targets,
                logits,
                self.top_k,
                self.name
            )

        return correct_predictions, accuracy_val, hits_at_k_val, mean_hits_at_k

    def build_output(
            self,
            hidden,
            hidden_size,
            regularizer=None,
            **kwargs
    ):
        output_tensors = {}

        # ================ Placeholder ================
        targets = self._get_output_placeholder()
        output_tensors[self.name] = targets
        logging.debug('  targets_placeholder: {0}'.format(targets))

        # ================ Predictions ================
        outs = self._get_predictions(
            hidden,
            hidden_size,
            regularizer=regularizer
        )
        (
            predictions,
            top_k_predictions,
            probabilities,
            logits,
            class_weights,
            class_biases
        ) = outs

        output_tensors[PREDICTIONS + '_' + self.name] = predictions
        output_tensors[TOP_K_PREDICTIONS + '_' + self.name] = top_k_predictions
        output_tensors[PROBABILITIES + '_' + self.name] = probabilities

        # ================ Measures ================
        correct_predictions, accuracy, hits_at_k, mean_hits_at_k = \
            self._get_measures(targets, predictions, logits)

        output_tensors[
            CORRECT_PREDICTIONS + '_' + self.name
            ] = correct_predictions
        output_tensors[ACCURACY + '_' + self.name] = accuracy
        output_tensors[HITS_AT_K + '_' + self.name] = hits_at_k
        output_tensors[MEAN_HITS_AT_K + '_' + self.name] = mean_hits_at_k

        if 'sampled' not in self.loss['type']:
            tf.summary.scalar(
                'train_batch_accuracy_{}'.format(self.name),
                accuracy
            )
            tf.summary.scalar(
                'train_batch_mean_hits_at_k_{}'.format(self.name),
                mean_hits_at_k
            )

        # ================ Loss (Cross Entropy) ================
        train_mean_loss, eval_loss = self._get_loss(
            targets,
            hidden,
            logits,
            probabilities,
            class_weights,
            class_biases
        )

        output_tensors[EVAL_LOSS + '_' + self.name] = eval_loss
        output_tensors[TRAIN_MEAN_LOSS + '_' + self.name] = train_mean_loss

        tf.summary.scalar(
            'train_mean_loss_{}'.format(self.name),
            train_mean_loss
        )

        return train_mean_loss, eval_loss, output_tensors

    @staticmethod
    def update_model_definition_with_metadata(
            output_feature,
            feature_metadata,
            *args,
            **kwargs
    ):
        output_feature['num_classes'] = feature_metadata['vocab_size']
        output_feature['top_k'] = min(
            output_feature['num_classes'],
            output_feature['top_k']
        )

        if isinstance(output_feature[LOSS]['class_weights'], (list, tuple)):
            output_feature[LOSS]['class_weights'] = (
                    [0] + output_feature[LOSS]['class_weights']  # for UNK
            )
            if (len(output_feature[LOSS]['class_weights']) !=
                    output_feature['num_classes']):
                raise ValueError(
                    'The length of class_weights ({}) is not compatible with '
                    'the number of classes ({})'.format(
                        len(output_feature[LOSS]['class_weights']),
                        output_feature['num_classes']
                    )
                )

        if output_feature[LOSS]['class_distance_temperature'] > 0:
            if 'distances' in feature_metadata:
                distances = feature_metadata['distances']
                temperature = output_feature[LOSS]['class_distance_temperature']

                for i in range(len(distances)):
                    distances[i, :] = softmax(
                        distances[i, :],
                        temperature=temperature
                    )

                output_feature[LOSS]['distances'] = distances
            else:
                raise ValueError(
                    'No class_distance metadata available for '
                    'feature {}'.format(output_feature['name'])
                )

        if output_feature[LOSS]['type'] == 'sampled_softmax_cross_entropy':
            output_feature[LOSS]['class_counts'] = [
                feature_metadata['str2freq'][cls]
                for cls in feature_metadata['idx2str']
            ]

    @staticmethod
    def calculate_overall_stats(
            test_stats,
            output_feature,
            dataset,
            train_set_metadata
    ):
        feature_name = output_feature['name']
        stats = test_stats[feature_name]
        confusion_matrix = ConfusionMatrix(
            dataset.get(feature_name),
            stats[PREDICTIONS],
            labels=train_set_metadata[feature_name]['idx2str']
        )
        stats['confusion_matrix'] = confusion_matrix.cm.tolist()
        stats['overall_stats'] = confusion_matrix.stats()
        stats['per_class_stats'] = confusion_matrix.per_class_stats()

    @staticmethod
    def postprocess_results(
            output_feature,
            result,
            metadata,
            experiment_dir_name,
            skip_save_unprocessed_output=False
    ):
        postprocessed = {}
        npy_filename = os.path.join(experiment_dir_name, '{}_{}.npy')
        name = output_feature['name']

        if PREDICTIONS in result and len(result[PREDICTIONS]) > 0:
            preds = result[PREDICTIONS]
            if 'idx2str' in metadata:
                postprocessed[PREDICTIONS] = [
                    metadata['idx2str'][pred] for pred in preds
                ]

            else:
                postprocessed[PREDICTIONS] = preds

            if not skip_save_unprocessed_output:
                np.save(npy_filename.format(name, PREDICTIONS), preds)

            del result[PREDICTIONS]

        if PROBABILITIES in result and len(result[PROBABILITIES]) > 0:
            probs = result[PROBABILITIES]
            prob = np.amax(probs, axis=1)
            postprocessed[PROBABILITIES] = probs
            postprocessed['probability'] = prob

            if not skip_save_unprocessed_output:
                np.save(npy_filename.format(name, PROBABILITIES), probs)
                np.save(npy_filename.format(name, 'probability'), probs)

            del result[PROBABILITIES]

        if ('predictions_top_k' in result and
            len(result['predictions_top_k'])) > 0:

            preds_top_k = result['predictions_top_k']
            if 'idx2str' in metadata:
                postprocessed['predictions_top_k'] = [
                    [metadata['idx2str'][pred] for pred in pred_top_k]
                    for pred_top_k in preds_top_k
                ]
            else:
                postprocessed['predictions_top_k'] = preds_top_k

            if not skip_save_unprocessed_output:
                np.save(
                    npy_filename.format(name, 'predictions_top_k'),
                    preds_top_k
                )

            del result['predictions_top_k']

        return postprocessed

    @staticmethod
    def populate_defaults(output_feature):
        set_default_value(
            output_feature,
            LOSS,
            {
                'type': 'softmax_cross_entropy',
                'sampler': None,
                'negative_samples': 0,
                'distortion': 1,
                'labels_smoothing': 0,
                'class_weights': 1,
                'robust_lambda': 0,
                'confidence_penalty': 0,
                'class_distance_temperature': 0,
                'weight': 1
            }
        )
        set_default_value(output_feature[LOSS], 'type', 'softmax_cross_entropy')

        if output_feature[LOSS]['type'] == 'sampled_softmax_cross_entropy':
            set_default_value(output_feature[LOSS], 'sampler', 'log_uniform')
            set_default_value(output_feature[LOSS], 'negative_samples', 25)
            set_default_value(output_feature[LOSS], 'distortion', 0.75)
        else:
            set_default_value(output_feature[LOSS], 'sampler', None)
            set_default_value(output_feature[LOSS], 'negative_samples', 0)
            set_default_value(output_feature[LOSS], 'distortion', 1)

        set_default_value(output_feature[LOSS], 'unique', False)
        set_default_value(output_feature[LOSS], 'weight', 1)

        set_default_value(output_feature, 'top_k', 3)
        set_default_value(output_feature, 'dependencies', [])
        set_default_value(output_feature, 'reduce_input', SUM)
        set_default_value(output_feature, 'reduce_dependencies', SUM)



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import argparse
import logging

import numpy as np
import pandas as pd

from ludwig.utils.data_utils import read_csv


def concatenate(train_csv, vali_csv, test_csv, output_csv):
    concatenated_df = concatenate_csv(train_csv, vali_csv, test_csv)

    logging.info('Saving concatenated csv..')
    concatenated_df.to_csv(output_csv, encoding='utf-8', index=False)
    logging.info('done')


def concatenate_csv(train_csv, vali_csv, test_csv):
    logging.info('Loading training csv...')
    train_df = read_csv(train_csv)
    logging.info('done')

    logging.info('Loading validation csv..')
    vali_df = read_csv(vali_csv) if vali_csv is not None else None
    logging.info('done')

    logging.info('Loading test csv..')
    test_df = read_csv(test_csv) if test_csv is not None else None
    logging.info('done')

    logging.info('Concatenating csvs..')
    concatenated_df = concatenate_df(train_df, vali_df, test_df)
    logging.info('done')

    return concatenated_df


def concatenate_df(train_df, vali_df, test_df):
    train_size = len(train_df)
    vali_size = len(vali_df) if vali_df is not None else 0
    test_size = len(test_df) if test_df is not None else 0
    concatenated_df = pd.concat([train_df, vali_df, test_df])
    split = np.array(
        [0] * train_size + [1] * vali_size + [2] * test_size,
        dtype=np.int8
    )
    concatenated_df = concatenated_df.assign(split=pd.Series(split).values)
    return concatenated_df


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Concatenate train validation and test set'
    )

    parser.add_argument(
        '-train',
        '--train_csv',
        help='CSV containing the training set'
    )
    parser.add_argument(
        '-vali',
        '--vali_csv',
        help='CSV containing the validation set'
    )
    parser.add_argument(
        '-test',
        '--test_csv',
        help='CSV containing the test set'
    )

    parser.add_argument('-o', '--output_csv', help='output csv')
    args = parser.parse_args()

    concatenate(args.train_csv, args.vali_csv, args.test_csv, args.output_csv)



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import h5py
import numpy as np


class Dataset:
    def __init__(self, dataset, input_features, output_features, data_hdf5_fp):
        self.dataset = dataset

        self.size = min(map(len, self.dataset.values()))

        self.input_features = {}
        for feature in input_features:
            feature_name = feature['name']
            self.input_features[feature_name] = feature
        self.output_features = {}
        for feature in output_features:
            feature_name = feature['name']
            self.output_features[feature_name] = feature
        self.features = self.input_features.copy()
        self.features.update(self.output_features)
        self.data_hdf5_fp = data_hdf5_fp

    def get(self, feature_name, idx=None):
        if idx is None:
            idx = range(self.size)
        if (self.data_hdf5_fp is None or
                'in_memory' not in self.features[feature_name]):
            return self.dataset[feature_name][idx]
        if self.features[feature_name]['in_memory']:
            return self.dataset[feature_name][idx]

        sub_batch = self.dataset[feature_name][idx]

        indices = np.empty((3, len(sub_batch)), dtype=np.int64)
        indices[0, :] = sub_batch
        indices[1, :] = np.arange(len(sub_batch))
        indices = indices[:, np.argsort(indices[0])]

        with h5py.File(self.data_hdf5_fp, 'r') as h5_file:
            im_data = h5_file[feature_name + '_data'][indices[0, :], :, :]
        indices[2, :] = np.arange(len(sub_batch))
        indices = indices[:, np.argsort(indices[1])]
        return im_data[indices[2, :]]

    def get_dataset(self):
        return self.dataset

    def set_dataset(self, dataset):
        self.dataset = dataset
